@@ -1,6 +1,6 @@
 import asyncio
 from . import tcp, udp
-from async_dns.core import TCP, DNSError, DNSMessage, NameServers, logger, types, REQUEST, Record, InvalidNameServer
+from async_dns.core import DNSError, DNSMessage, NameServers, logger, types, REQUEST, Record
 
 A_TYPES = types.A, types.AAAA
 PENDING = 0
@@ -8,40 +8,33 @@ RESOLVED = 1
 REJECTED = 2
 
 class Query:
-    def __init__(self, resolver, loop, fqdn, qtype, tick):
-        self.loop = loop
+    protocols = {
+        'tcp': tcp.request,
+        'udp': udp.request,
+    }
+
+    def __init__(self, resolver, fqdn, qtype, tick):
         self.resolver = resolver
         self.fqdn = fqdn
         self.qtype = qtype
         self.tick = tick
-        self.future = loop.create_future()
-        self.from_cache = False
+        self._cached = False
         self._status = PENDING
         self._result = DNSMessage(ra=resolver.recursive)
         self._result.qd.append(Record(REQUEST, name=fqdn, qtype=qtype))
 
     async def query(self):
-        error = None
         domain = self.fqdn
         nameservers = None
-        try:
-            while True:
-                if not self.future.cancelled():
-                    self.from_cache = await self.query_cache(domain)
-                    logger.debug('[query_cache][%s][%s] %s', types.get_name(self.qtype), domain, self.from_cache)
-                    if self.from_cache: break
-                if not self.future.cancelled():
-                    remote_res = await self.query_remote(domain, nameservers)
-                    logger.debug('[query_remote][%s][%s] %s', types.get_name(self.qtype), domain, remote_res)
-                    if remote_res is None: break
-                    domain, nameservers = remote_res
-        except Exception as e:
-            error = e
-        if not self.future.cancelled():
-            if error:
-                self.future.set_exception(error)
-            else:
-                self.future.set_result((self._result, self.from_cache))
+        while True:
+            self._cached = await self.query_cache(domain)
+            logger.debug('[query_cache][%s][%s] %s', types.get_name(self.qtype), domain, self._cached)
+            if self._cached: break
+            remote_res = await self.query_remote(domain, nameservers)
+            logger.debug('[query_remote][%s][%s] %s', types.get_name(self.qtype), domain, remote_res)
+            if remote_res is None: break
+            domain, nameservers = remote_res
+        return self._result, self._cached
 
     async def query_cache(self, domain):
         '''Returns a boolean whether a cache hit occurs.'''
@@ -57,7 +50,7 @@ class Query:
             if self.qtype == types.CNAME:
                 return True
             for rec in cname:
-                inter_res = await resolver.query(rec.data, self.qtype)
+                inter_res = await resolver.query_safe(rec.data, self.qtype)
                 if inter_res is None or inter_res.r > 0:
                     continue
                 result.an.extend(inter_res.an)
@@ -157,36 +150,38 @@ class Query:
         return inter_res
 
     async def request_remote(self, nameservers, req):
-        while not self.future.cancelled():
-            addr = nameservers.get()
-            if not addr:
-                raise InvalidNameServer
+        last_err = None
+        for addr in nameservers.iter():
             try:
-                data = await self.request_once(req, addr)
-                inter_res = DNSMessage.parse(data)
+                inter_res = await self.request_once(req, addr)
                 logger.debug('[request_remote] %s', inter_res)
                 if inter_res.qd[0].name != req.qd[0].name:
                     raise DNSError(-1, 'Question section mismatch')
                 assert inter_res.r != 2, 'Remote server fail'
-            except (asyncio.TimeoutError, AssertionError) as e:
-                logger.debug('[request_remote][server_error] %s', e)
-            except DNSError as e:
-                logger.debug('[request_remote][dns_error] %s', e)
             except Exception as e:
-                logger.error('[request_remote][error] %s', e)
+                error_type = 'error'
+                if isinstance(e, (asyncio.TimeoutError, AssertionError)):
+                    error_type = 'server_error'
+                elif isinstance(e, DNSError):
+                    error_type = 'dns_error'
+                else:
+                    error_type = 'error'
+                    import traceback
+                    traceback.print_exc()
+                logger.debug('[request_remote][%s] %s %s', error_type, str(addr), repr(e))
+                last_err = e
             else:
                 nameservers.success(addr)
                 return inter_res
             nameservers.fail(addr)
+        else:
+            raise last_err
 
     async def request_once(self, req, addr):
         '''Return response to a request.
 
         Send DNS request data with `protocol`.
         '''
-        if addr.protocol is TCP:
-            request = tcp.request
-        else:
-            request = udp.request
+        request = self.protocols[addr.protocol]
         data = await request(req, addr, self.resolver.request_timeout)
         return data
